%% Camera tracking using direct image alignment
%clear all; close all; clc;

data_dir = '/Users/Rui/Data/rgbd_dataset_freiburg2_desk';
match_file_path = 'matches_fr2_desk.txt';
K = [520.9 0 325.1;	0 521.0 249.7; 0 0 1]; %fr2
%K = [525.0 0 319.5; 0 525.0	239.5; 0 0 2]; % ROS default

% data_dir = '/Users/Rui/Data/rgbd_dataset_freiburg1_desk';
% match_file_path = '../matches_fr1_desk.txt';
% K = [517.3 0 318.6; 0 516.5	255.3; 0 0 1]; % fr1

% Read files names
fid = fopen(match_file_path);
nb = 0;
filenames = cell(3000, 2);

while ~feof(fid)
    nb = nb + 1;
    line = fgetl(fid); 
    filenames(nb, 1:2) = textscan(line,'%s %s'); 
end

fclose(fid);
filenames = filenames(1:nb, :);

height = 480;
width = 640;

% Pre-allocate memory
% Relative poses to reference keyframe
Rs_kc = zeros(3, 3, nb);
Ts_kc = zeros(3, nb);

% Reference frame id
ref_ids = zeros(nb, 1);

% Only store images for KFs
kf_rgb_imgs = zeros(height, width, nb);
kf_depth_imgs = zeros(height, width, nb);

% Global poses of KFs
kf_Rs_wk = zeros(3, 3, nb);
kf_Ts_wk = zeros(3, nb);
kf_ids = zeros(nb, 1);

% Set first frame
kf_Rs_wk(:, :, 1) = eye(3);
kf_Ts_wk(:, 1) = zeros(3, 1);
kf_rgb_imgs(:, :, 1) = double(imreadbw([data_dir '/' filenames{1, 1}{1}]));
kf_depth_imgs(:, :, 1) = double(imread([data_dir '/' filenames{1, 2}{1}]))/5000;
ref_ids(1) = 1;
Rs_kc(:, :, 1) = eye(3);
Ts_kc(:, 1) = zeros(3, 1);
kf_ids(1) = 1;

huber_delta = 0.2;
curr_kf_id = 1;
start_level = 5;
end_level = 2;
threshold = 0.003;
kf_num = 1;

for i = 2:nb
    fprintf('Track frame: %d\n', i);
    
    % Read new image
    rgb_img = double(imreadbw([data_dir '/' filenames{i, 1}{1}]));
    depth_img = double(imread([data_dir '/' filenames{i, 2}{1}]))/5000;
    
    % initialization
    xi = [0 0 0 0 0 0]';
    
    % pyramid levels
    for lvl = start_level:-1:end_level
        % get downscaled image, depth image, and K-matrix of down-scaled image.
        [IRef, DRef, Klvl] = downscale(kf_rgb_imgs(:, :, kf_num),...
            kf_depth_imgs(:, :, kf_num), K, lvl);
        [I, D] = downscale(rgb_img, depth_img, K, lvl);
        
        % just do at most 20 steps.
        errLast = 1e10;
        for iter=1:20
            % calculate Jacobian of residual function (Matrix of dim (width*height) x 6)
            [Jac, residual, weights] = deriveResidualsAnalytic(IRef,...
                DRef, I, xi, Klvl, huber_delta);
            
            % set rows with NaN to 0 (e.g. because out-of-bounds or invalid depth).
            notValid = isnan(sum(Jac,2)+residual);
            residual(notValid,:) = 0;
            Jac(notValid,:) = 0;
            weights(notValid,:) = 0;

            % do Gauss-Newton step
            upd = - (Jac' * (repmat( weights, 1, 6 ) .* Jac))^-1 * Jac'...
                * (weights .* residual);
            
            % multiply increment from left onto the current estimate
            lastXi = xi;
            xi = se3Log(se3Exp(upd) * se3Exp(xi));
            
            err = mean(residual .* residual);
            
            if(err / errLast > 0.995)
                break;
            end   
            
            errLast = err;
        end
    end
    
    ref_ids(i) = kf_num;
    
    % Store results
    T_kc = inv(se3Exp(xi));
    Rs_kc(:, :, i) = T_kc(1:3, 1:3);
    Ts_kc(:, i) = T_kc(1:3, 4);
    
    %xi = se3Log(inv([Rs_kc(:, :, i) Ts_kc(:, i); 0 0 0 1]));
    
    % Check if new KF is needed
    dist = xi' * xi;
    if (dist > threshold)
        kf_num = kf_num + 1;
        
        % Store images
        kf_rgb_imgs(:, :, kf_num) = rgb_img;
        kf_depth_imgs(:, :, kf_num) = depth_img;
        
        % Calculate and store global pose
        T_wk = [kf_Rs_wk(:, :, ref_ids(i)) kf_Ts_wk(:, ref_ids(i));...
            0 0 0 1] * T_kc;
        kf_Rs_wk(:, :, kf_num) = T_wk(1:3, 1:3);
        kf_Ts_wk(:, kf_num) = T_wk(1:3, 4);

        kf_ids(kf_num) = i;
    end
end

kf_rgb_imgs = kf_rgb_imgs(:, :, 1:kf_num);
kf_depth_imgs = kf_depth_imgs(:, :, 1:kf_num);
kf_Rs_wk = kf_Rs_wk(:, :, 1:kf_num);
kf_Ts_wk = kf_Ts_wk(:, 1:kf_num);
kf_ids = kf_ids(1:kf_num);

% save('fr2_desk_5_2_Rs_kc.mat', 'Rs_kc');
% save('fr2_desk_5_2_Ts_kc.mat', 'Ts_kc');
% save('fr2_desk_5_2_ref_ids.mat', 'ref_ids');
% save('fr2_desk_5_2_kf_Rs_wk.mat', 'kf_Rs_wk');
% save('fr2_desk_5_2_kf_Ts_wk.mat', 'kf_Ts_wk');
% save('fr2_desk_5_2_filenames.mat', 'filenames');
% save('fr2_desk_5_2_kf_rgb_imgs.mat', 'kf_rgb_imgs');
% save('fr2_desk_5_2_kf_depth_imgs.mat', 'kf_depth_imgs');
% save('fr2_desk_5_2_kf_ids.mat', 'kf_ids');

% Output file for evaluation
fname = sprintf('fr2_desk_level_%d_%d_sampled.txt', start_level, end_level);
fid = fopen(fname, 'w');
for i = 1:1:nb
    ref_id = ref_ids(i);
    T_wk = [kf_Rs_wk(:, :, ref_id) kf_Ts_wk(:, ref_id); 0 0 0 1];
    T_kc = [Rs_kc(:, :, i) Ts_kc(:, i); 0 0 0 1];
    T_wc = T_wk * T_kc;
    quat = rotm2quat(T_wc(1:3, 1:3));
    timestamp = textscan(filenames{i, 2}{1}, 'depth/%f.png');
    fprintf(fid,'%.6f %f %f %f %f %f %f %f\n', timestamp{1}, T_wc(1, 4),...
        T_wc(2, 4),T_wc(3, 4), quat(1), quat(2), quat(3), quat(4));
end
fclose(fid);

%% Pose-graph optimization

% Load saved data
% files = dir('*.mat'); for i = 1:numel(files) load(files(i).name); end

% Manually detect loop
% 1-2563 (kf_id: 1-392) % id in match file!!!
ref_id = 392;
curr_id = 1;  

% Estimate relative pose using direct image alignment    
% initialization
xi_loop = [0 0 0 0 0 0]';
% pyramid levels
for lvl = 5:-1:1
    % get downscaled image, depth image, and K-matrix of down-scaled image.
    [IRef, DRef, Klvl] = downscale(kf_rgb_imgs(:, :, ref_id),...
        kf_depth_imgs(:, :, ref_id), K, lvl);
    [I, D] = downscale(kf_rgb_imgs(:, :, curr_id), ...
        kf_depth_imgs(:, :, curr_id), K, lvl);
    
    % just do at most 20 steps.
    errLast = 1e10;
    for iter=1:20
        % calculate Jacobian of residual function (Matrix of dim (width*height) x 6)
        [Jac, residual, weights] = deriveResidualsAnalytic(IRef,...
            DRef, I, xi_loop, Klvl, huber_delta);
        
        % set rows with NaN to 0 (e.g. because out-of-bounds or invalid depth).
        notValid = isnan(sum(Jac,2)+residual);
        residual(notValid,:) = 0;
        Jac(notValid,:) = 0;
        weights(notValid,:) = 0;
        
        % do Gauss-Newton step
        upd = - (Jac' * (repmat( weights, 1, 6 ) .* Jac))^-1 * Jac'...
            * (weights .* residual);
        
        % multiply increment from left onto the current estimate
        xi_loop = se3Log(se3Exp(upd) * se3Exp(xi_loop));
        
        err = mean(residual .* residual);
        
        if(err / errLast > 0.995)
            break;
        end
        
        errLast = err;
    end
end

% Show residual
figure;
imagesc(reshape(residual,size(kf_rgb_imgs(:, :, curr_id))));
colormap gray;
set(gca, 'CLim', [-1,1]);

% Plot trajectory before loop closure
figure;
plot3(kf_Ts_wk(1, :), kf_Ts_wk(2, :), kf_Ts_wk(3, :), 'b');
axis equal;
xlabel('x');
ylabel('y');
zlabel('z');
hold on;

% Gauss-Newton
% Initial state
Xi = zeros(kf_num * 6, 1);
for i = 1:kf_num
   xi = se3Log([kf_Rs_wk(:, :, i) kf_Ts_wk(:, i); 0 0 0 1]);
   Xi(6 * i - 5:6 * i) = xi;
end

kf_Ts_wk_new = zeros(3, kf_num);
kf_Rs_wk_new = zeros(3, 3, kf_num);

errLast = 1e10;
for iter = 1:20
    fprintf('Loop closure iteration: %d\n', iter);
    % calculate Jacobian of residual function
    % Calculate residual
    Res = zeros(kf_num * 6, 1);
    for i = 1:kf_num - 1
        T_ji = inv([Rs_kc(:, :, kf_ids(i+1)) Ts_kc(:, kf_ids(i+1)); 0 0 0 1]);
        T_wi = se3Exp(Xi(6 * i - 5:6 * i));
        T_wj = se3Exp(Xi(6 * i + 1:6 * i + 6));
        Res(6 * i - 5:6 * i) = se3Log(inv(T_wi) * T_wj * T_ji);
    end
    % The part for loop contraints
    T_ji = se3Exp(xi_loop);
    T_wi = se3Exp(Xi(6 * ref_id - 5:6 * ref_id));
    T_wj = se3Exp(Xi(6 * curr_id - 5:6 * curr_id));
    Res(end - 5:end) = se3Log(inv(T_wi) * T_wj * T_ji);
    
    % Build up Jacobian
    J = zeros(kf_num * 6);
    for i = 1:kf_num - 1
        T_wi = se3Exp(Xi(6 * i - 5:6 * i));
        Ad = adjoint(inv(T_wi));
        J(6 * i - 5:6 * i, 6 * i - 5:6 * i) = -Ad;
        J(6 * i - 5:6 * i, 6 * i + 1:6 * i + 6) = Ad;
    end
    % Jacobian for the loop constraint
    T_wi = se3Exp(Xi(6 * ref_id - 5:6 * ref_id));
    Ad = adjoint(inv(T_wi));
    J(end - 5:end, 6 * curr_id - 5:6 * curr_id) = Ad;
    J(end - 5:end, 6 * ref_id - 5:6 * ref_id) = -Ad;
    
    % Fix the first KF
    J = J(:, 7:end);
    
    % do Gauss-Newton step
    upd = - inv(J' * J) * J' * Res;
    
    % multiply increment from left onto the current estimate
    for i = 1:kf_num - 1
        delta = upd(6 * i - 5:6 * i);
        xi = Xi(6 * i + 1:6 * i + 6);
        T_wi_new = se3Exp(delta) * se3Exp(xi);
        Xi(6 * i + 1:6 * i + 6) = se3Log(T_wi_new);
        kf_Ts_wk_new(:, i + 1) = T_wi_new(1:3, 4);
        kf_Rs_wk_new(:, :, i + 1) = T_wi_new(1:3, 1:3);
    end   
    
    err = Res' * Res
    
    if(err / errLast > 0.995)
        break;
    end
    
    errLast = err;
end

plot3(kf_Ts_wk_new(1, :), kf_Ts_wk_new(2, :), kf_Ts_wk_new(3, :), 'r');

% Output file for evaluation
fname = sprintf('fr2_desk_level_%d_%d_loopclosure.txt', start_level, end_level);
fid = fopen(fname, 'w');
for i = 1:1:nb
    ref_id = ref_ids(i);
    T_wk = [kf_Rs_wk_new(:, :, ref_id) kf_Ts_wk_new(:, ref_id); 0 0 0 1];
    T_kc = [Rs_kc(:, :, i) Ts_kc(:, i); 0 0 0 1];
    T_wc = T_wk * T_kc;
    quat = rotm2quat(T_wc(1:3, 1:3));
    timestamp = textscan(filenames{i, 2}{1}, 'depth/%f.png');
    fprintf(fid,'%.6f %f %f %f %f %f %f %f\n', timestamp{1}, T_wc(1, 4),...
        T_wc(2, 4),T_wc(3, 4), quat(1), quat(2), quat(3), quat(4));
end
fclose(fid);








